from __future__ import annotations

from typing import TYPE_CHECKING

from django.db import models

from ...subject_schedule import SubjectSchedule

if TYPE_CHECKING:
    from ...schedule import Schedule
    from ...visit_schedule import VisitSchedule


class CrfScheduleModelMixin(models.Model):
    """A mixin for CRF models to add the ability to determine
    if the subject is on/off schedule.

    To be declared with VisitMethodsModelMixin to get access
    to `related_visit` and `subject_identifier`.
    """

    # If True, compares report_datetime and offschedule_datetime as datetimes
    # If False, (Default) compares report_datetime and
    # offschedule_datetime as dates
    offschedule_compare_dates_as_datetimes = False

    def save(self, *args, **kwargs):
        self.is_onschedule_or_raise()
        super().save(*args, **kwargs)

    @property
    def visit_schedule_name(self) -> str:
        return self.related_visit.visit_schedule_name

    @property
    def schedule_name(self) -> str:
        return self.related_visit.schedule_name

    @property
    def visit_schedule(self) -> VisitSchedule:
        return self.related_visit.visit_schedule

    @property
    def schedule(self) -> Schedule:
        return self.related_visit.schedule

    def is_onschedule_or_raise(self) -> None:
        subject_schedule = SubjectSchedule(
            self.subject_identifier,
            visit_schedule=self.visit_schedule,
            schedule=self.schedule,
        )
        subject_schedule.onschedule_or_raise(
            report_datetime=self.report_datetime,
            compare_as_datetimes=self.offschedule_compare_dates_as_datetimes,
        )

    class Meta:
        abstract = True



from django.apps import apps as django_apps
from django.db import models
from django.db.models import Q, UniqueConstraint
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel
from edc_model.validators import datetime_not_future
from edc_protocol.validators import datetime_not_before_study_start
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin
from edc_utils import get_utcnow

from ..choices import SCHEDULE_STATUS
from ..model_mixins import VisitScheduleFieldsModelMixin


class OnScheduleModelError(Exception):
    pass


class SubjectScheduleModelManager(models.Manager):
    def get_by_natural_key(self, subject_identifier, visit_schedule_name, schedule_name):
        return self.get(
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
        )

    def onschedules(self, subject_identifier=None, report_datetime=None):
        """Returns a list of onschedule model instances for this
        subject where the schedule_status would be ON_SCHEDULE
        relative to the report_datetime.
        """
        onschedules = []
        report_datetime = report_datetime or get_utcnow()
        qs = self.filter(
            Q(subject_identifier=subject_identifier),
            Q(onschedule_datetime__lte=report_datetime),
            (
                Q(offschedule_datetime__gte=report_datetime)
                | Q(offschedule_datetime__isnull=True)
            ),
        )
        for obj in qs:
            onschedule_model_cls = django_apps.get_model(obj.onschedule_model)
            onschedules.append(
                onschedule_model_cls.objects.get(subject_identifier=subject_identifier)
            )
        return onschedules


class SubjectScheduleHistory(
    NonUniqueSubjectIdentifierFieldMixin,
    VisitScheduleFieldsModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    onschedule_model = models.CharField(max_length=100)

    offschedule_model = models.CharField(max_length=100)

    onschedule_datetime = models.DateTimeField(
        validators=[datetime_not_before_study_start, datetime_not_future]
    )

    offschedule_datetime = models.DateTimeField(
        validators=[datetime_not_before_study_start, datetime_not_future], null=True
    )

    schedule_status = models.CharField(max_length=15, choices=SCHEDULE_STATUS, null=True)

    objects = SubjectScheduleModelManager()

    on_site = CurrentSiteManager()

    def natural_key(self):
        return (
            self.subject_identifier,
            self.visit_schedule_name,
            self.schedule_name,
        )

    class Meta(BaseUuidModel.Meta, NonUniqueSubjectIdentifierFieldMixin.Meta):
        constraints = [
            UniqueConstraint(
                fields=["subject_identifier", "visit_schedule_name", "schedule_name"],
                name="%(app_label)s_%(class)s_subject_uniq",
            )
        ]
        indexes = (
            BaseUuidModel.Meta.indexes + NonUniqueSubjectIdentifierFieldMixin.Meta.indexes
        )



from .crf import CrfScheduleModelMixin
from .off_schedule_model_mixin import OffScheduleModelMixin
from .on_schedule_model_mixin import OnScheduleModelMixin
from .subject_on_schedule_model_mixin import SubjectOnScheduleModelMixin
from .visit_schedule import (
    VisitCodeFieldsModelMixin,
    VisitScheduleFieldsModelMixin,
    VisitScheduleMethodsModelMixin,
    VisitScheduleModelMixin,
    VisitScheduleModelMixinError,
)



from dateutil.relativedelta import relativedelta
from django.apps import apps as django_apps
from django.test import TestCase

from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.system_checks import (
    check_form_collections,
    visit_schedule_check,
)
from edc_visit_schedule.visit import CrfCollection, FormsCollectionError, Visit
from edc_visit_schedule.visit.crf import Crf
from edc_visit_schedule.visit_schedule import VisitSchedule
from visit_schedule_app.consents import consent_v1


class TestSystemChecks(TestCase):
    def setUp(self):
        self.visit_schedule = VisitSchedule(
            name="visit_schedule",
            verbose_name="Visit Schedule",
            offstudy_model="visit_schedule_app.subjectoffstudy",
            death_report_model="visit_schedule_app.deathreport",
            locator_model="edc_locator.subjectlocator",
        )

        self.schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            appointment_model="edc_appointment.appointment",
            consent_definitions=[consent_v1],
            base_timepoint=1,
        )

    def test_system_check(self):
        site_visit_schedules._registry = {}
        errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "edc_visit_schedule.001")

    def test_visit_schedule_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)
        errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(errors), 0)

    def test_visit_schedule_bad_model(self):
        site_visit_schedules._registry = {}
        visit_schedule = VisitSchedule(
            name="visit_schedule",
            verbose_name="Visit Schedule",
            offstudy_model="blah.subjectoffstudy",
            death_report_model="visit_schedule_app.deathreport",
            locator_model="edc_locator.subjectlocator",
        )
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            appointment_model="edc_appointment.appointment",
            consent_definitions=[consent_v1],
            base_timepoint=1,
        )
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)
        errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(errors), 1)
        self.assertEqual("edc_visit_schedule.visit_schedules", errors[0].id)

    def test_schedule_bad_model(self):
        site_visit_schedules._registry = {}
        visit_schedule = VisitSchedule(
            name="visit_schedule",
            verbose_name="Visit Schedule",
            offstudy_model="visit_schedule_app.subjectoffstudy",
            death_report_model="visit_schedule_app.deathreport",
            locator_model="edc_locator.subjectlocator",
        )
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            appointment_model="blah.appointment",
            consent_definitions=[consent_v1],
            base_timepoint=1,
        )
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)
        errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(errors), 1)
        self.assertEqual("edc_visit_schedule.schedules", errors[0].id)

    def test_schedule_bad_crf_model(self):
        site_visit_schedules._registry = {}
        visit_schedule = VisitSchedule(
            name="visit_schedule",
            verbose_name="Visit Schedule",
            offstudy_model="visit_schedule_app.subjectoffstudy",
            death_report_model="visit_schedule_app.deathreport",
            locator_model="edc_locator.subjectlocator",
        )
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            appointment_model="edc_appointment.appointment",
            consent_definitions=[consent_v1],
            base_timepoint=1,
        )
        crfs = CrfCollection(
            Crf(show_order=10, model="blah.CrfOne"),
            Crf(show_order=20, model="blah.CrfTwo"),
            Crf(show_order=30, model="blah.CrfThree"),
        )
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)
        errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(errors), 3)
        self.assertEqual("edc_visit_schedule.visits", errors[0].id)

    def test_visit_with_crfs_and_prns_and_unscheduled_and_missed_crfs_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne"),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
            Crf(show_order=30, model="visit_schedule_app.CrfThree"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.PrnOne"),
            Crf(show_order=25, model="visit_schedule_app.PrnTwo"),
            Crf(show_order=35, model="visit_schedule_app.PrnThree"),
        )
        unscheduled_crfs = CrfCollection(
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
            Crf(show_order=30, model="visit_schedule_app.CrfThree"),
        )
        missed_crfs = CrfCollection(Crf(show_order=30, model="visit_schedule_app.CrfThree"))

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            crfs_unscheduled=unscheduled_crfs,
            crfs_missed=missed_crfs,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_crf_not_required_with_matching_prn_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne", required=False),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
            Crf(show_order=30, model="visit_schedule_app.CrfThree"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOne"),
        )
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_proxy_crf_not_required_with_matching_proxy_prn_ok(
        self,
    ):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=20, model="visit_schedule_app.CrfTwoProxyOne", required=False),
        )
        prns = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfTwoProxyOne"),
        )
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_proxy_crf_not_required_without_shared_proxy_root_and_differing_proxy_prn_raises(
        self,
    ):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=20, model="visit_schedule_app.CrfTwoProxyOne", required=False),
        )
        prns = CrfCollection(
            Crf(show_order=25, model="visit_schedule_app.CrfTwoProxyOne"),
            Crf(show_order=15, model="visit_schedule_app.CrfTwoProxyTwo"),
        )
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 3)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            "Multiple proxies with same proxy root model appear for a visit. ",
            fc_errors[0].msg,
        )
        self.assertIn("Scheduled", fc_errors[0].msg)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crftwo": [
                        "visit_schedule_app.crftwoproxyone",
                        "visit_schedule_app.crftwoproxyone",
                        "visit_schedule_app.crftwoproxytwo",
                    ]
                }
            ),
            fc_errors[0].msg,
        )
        # Expect only the PRN fields to be flagged for the (undefined)
        # unscheduled and unscheduled visits
        self.assertEqual("edc_visit_schedule.004", fc_errors[1].id)
        self.assertIn("Unscheduled", fc_errors[1].msg)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crftwo": [
                        "visit_schedule_app.crftwoproxyone",
                        "visit_schedule_app.crftwoproxytwo",
                    ]
                }
            ),
            fc_errors[1].msg,
        )
        self.assertIn("Missed", fc_errors[2].msg)
        self.assertEqual("edc_visit_schedule.004", fc_errors[2].id)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crftwo": [
                        "visit_schedule_app.crftwoproxyone",
                        "visit_schedule_app.crftwoproxytwo",
                    ]
                }
            ),
            fc_errors[2].msg,
        )

    def test_proxy_crf_not_required_with_shared_proxy_root_and_matching_proxy_prn_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=20,
                model="visit_schedule_app.CrfTwoProxyOne",
                required=False,
                shares_proxy_root=True,
            ),
        )
        prns = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfTwoProxyOne",
                shares_proxy_root=True,
            ),
        )
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_duplicates_in_crf_collection_raises(self):
        """Duplicate crf/proxy in single collection = BAD."""
        # CRFs
        with self.assertRaises(FormsCollectionError) as cm:
            CrfCollection(
                Crf(show_order=10, model="visit_schedule_app.CrfOne"),
                Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
                Crf(show_order=30, model="visit_schedule_app.CrfTwo"),
            )
        self.assertIn(
            "Expected to be a unique sequence of crf/models.",
            str(cm.exception),
        )
        self.assertIn(
            "Duplicates ['visit_schedule_app.crftwo'].",
            str(cm.exception),
        )

        # Proxy models
        with self.assertRaises(FormsCollectionError) as cm:
            CrfCollection(
                Crf(show_order=10, model="visit_schedule_app.CrfOne"),
                Crf(show_order=20, model="visit_schedule_app.CrfTwoProxyOne"),
                Crf(show_order=30, model="visit_schedule_app.CrfTwoProxyOne"),
            )
        self.assertIn(
            "Expected to be a unique sequence of crf/models.",
            str(cm.exception),
        )
        self.assertIn(
            "Duplicates ['visit_schedule_app.crftwoproxyone'].",
            str(cm.exception),
        )

        # PRNs
        with self.assertRaises(FormsCollectionError) as cm:
            CrfCollection(
                Crf(show_order=15, model="visit_schedule_app.PrnOne"),
                Crf(show_order=25, model="visit_schedule_app.PrnTwo"),
                Crf(show_order=35, model="visit_schedule_app.PrnTwo"),
            )
        self.assertIn(
            "Expected to be a unique sequence of crf/models.",
            str(cm.exception),
        )
        self.assertIn(
            "Duplicates ['visit_schedule_app.prntwo'].",
            str(cm.exception),
        )

    def test_required_crf_in_prns_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne", required=True),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
            Crf(show_order=30, model="visit_schedule_app.CrfThree"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOne"),
            Crf(show_order=25, model="visit_schedule_app.PrnOne"),
            Crf(show_order=35, model="visit_schedule_app.PrnTwo"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_required_proxy_crf_in_prns_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                required=True,
                shares_proxy_root=True,
            ),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
        )
        prns = CrfCollection(
            Crf(
                show_order=15,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            ),
            Crf(show_order=25, model="visit_schedule_app.PrnOne"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_required_proxy_root_crf_with_required_child_proxy_crf_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne", required=True),
            Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne", required=True),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
            Crf(show_order=30, model="visit_schedule_app.CrfThree"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.003", fc_errors[0].id)
        self.assertIn(
            "Proxy root model class appears alongside associated child proxy for a visit.",
            fc_errors[0].msg,
        )
        self.assertIn("proxy_root_model=visit_schedule_app.crfone", fc_errors[0].msg)
        self.assertIn("proxy_model=visit_schedule_app.crfoneproxyone", fc_errors[0].msg)

    def test_required_proxy_root_crf_with_not_required_child_proxy_crf_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne", required=True),
            Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne", required=False),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
            Crf(show_order=30, model="visit_schedule_app.CrfThree"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.003", fc_errors[0].id)
        self.assertIn(
            "Proxy root model class appears alongside associated child proxy for a visit.",
            fc_errors[0].msg,
        )
        self.assertIn("proxy_root_model=visit_schedule_app.crfone", fc_errors[0].msg)
        self.assertIn("proxy_model=visit_schedule_app.crfoneproxyone", fc_errors[0].msg)

    def test_required_proxy_root_crf_with_child_proxy_prn_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne", required=True),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne"),
            Crf(show_order=25, model="visit_schedule_app.PrnOne"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.003", fc_errors[0].id)
        self.assertIn(
            "Proxy root model class appears alongside associated child proxy for a visit.",
            fc_errors[0].msg,
        )
        self.assertIn("proxy_root_model=visit_schedule_app.crfone", fc_errors[0].msg)
        self.assertIn("proxy_model=visit_schedule_app.crfoneproxyone", fc_errors[0].msg)

    def test_required_proxy_child_crf_with_proxy_root_prn_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOneProxyOne", required=True),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOne"),
            Crf(show_order=25, model="visit_schedule_app.PrnOne"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.003", fc_errors[0].id)
        self.assertIn(
            "Proxy root model class appears alongside associated child proxy for a visit.",
            fc_errors[0].msg,
        )
        self.assertIn("proxy_root_model=visit_schedule_app.crfone", fc_errors[0].msg)
        self.assertIn("proxy_model=visit_schedule_app.crfoneproxyone", fc_errors[0].msg)

    def test_not_required_crf_also_in_prns_ok(self):
        """e.g. CRF not required (toggled by metadata rule) but
        also available as optional PRN.
        """
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne", required=False),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOne"),
            Crf(show_order=25, model="visit_schedule_app.PrnOne"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_not_required_proxy_crf_with_shared_proxy_root_also_in_prns_ok(self):
        """e.g. Proxy CRF not required (toggled by metadata rule) but
        also available as optional PRN.
        """
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                required=False,
                shares_proxy_root=True,
            ),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
        )
        prns = CrfCollection(
            Crf(
                show_order=15,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            ),
            Crf(show_order=25, model="visit_schedule_app.PrnOne"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_three_crf_proxy_with_shares_root_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            ),
            Crf(
                show_order=20,
                model="visit_schedule_app.CrfOneProxyTwo",
                shares_proxy_root=True,
            ),
            Crf(
                show_order=30,
                model="visit_schedule_app.CrfOneProxyThree",
                shares_proxy_root=True,
            ),
        )
        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_crf_and_prn_proxies_with_shared_root_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                required=False,
                shares_proxy_root=True,
            ),
            Crf(
                show_order=20,
                model="visit_schedule_app.CrfOneProxyTwo",
                required=False,
                shares_proxy_root=True,
            ),
        )
        prns = CrfCollection(
            Crf(
                show_order=35,
                model="visit_schedule_app.CrfOneProxyThree",
                shares_proxy_root=True,
            ),
            Crf(show_order=25, model="visit_schedule_app.PrnOne"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_one_crf_proxy_shares_root_true_with_one_proxy_not_defined_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            ),
        )
        prns = CrfCollection(
            Crf(show_order=35, model="visit_schedule_app.CrfOneProxyThree"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            "Multiple proxies with same proxy root model appear for a visit.",
            fc_errors[0].msg,
        )
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[0].msg,
        )

    def test_two_crf_proxy_shares_root_true_with_one_proxy_not_defined_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            ),
            Crf(
                show_order=20,
                model="visit_schedule_app.CrfOneProxyTwo",
            ),
        )
        prns = CrfCollection(
            Crf(
                show_order=35,
                model="visit_schedule_app.CrfOneProxyThree",
                shares_proxy_root=True,
            ),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            "Multiple proxies with same proxy root model appear for a visit.",
            fc_errors[0].msg,
        )
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                        "visit_schedule_app.crfoneproxytwo",
                    ]
                }
            ),
            fc_errors[0].msg,
        )

    def test_one_crf_proxy_shares_root_true_with_two_proxy_not_defined_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOneProxyOne"),
            Crf(show_order=20, model="visit_schedule_app.CrfOneProxyTwo"),
        )
        prns = CrfCollection(
            Crf(
                show_order=35,
                model="visit_schedule_app.CrfOneProxyThree",
                shares_proxy_root=True,
            ),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            "Multiple proxies with same proxy root model appear for a visit.",
            fc_errors[0].msg,
        )
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                        "visit_schedule_app.crfoneproxytwo",
                    ]
                }
            ),
            fc_errors[0].msg,
        )

    def test_proxy_roots_shared_but_not_defined_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOneProxyOne"),
            Crf(show_order=12, model="visit_schedule_app.CrfOneProxyTwo"),
            Crf(show_order=20, model="visit_schedule_app.CrfTwoProxyOne"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne"),
            Crf(show_order=25, model="visit_schedule_app.CrfTwoProxyTwo"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxytwo",
                    ],
                    "visit_schedule_app.crftwo": [
                        "visit_schedule_app.crftwoproxyone",
                        "visit_schedule_app.crftwoproxytwo",
                    ],
                }
            ),
            fc_errors[0].msg,
        )

    def test_proxy_roots_shared_but_not_defined_case2(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOneProxyOne"),
            Crf(show_order=12, model="visit_schedule_app.CrfOneProxyTwo"),
            Crf(show_order=20, model="visit_schedule_app.CrfTwoProxyOne"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne"),
            Crf(show_order=25, model="visit_schedule_app.CrfTwoProxyOne"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxytwo",
                    ]
                }
            ),
            fc_errors[0].msg,
        )

    def test_proxy_pair_plus_another_proxy_with_same_parent_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOneProxyOne"),
            Crf(show_order=25, model="visit_schedule_app.CrfOneProxyTwo"),
        )
        prns = CrfCollection(Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne"))

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxytwo",
                    ]
                }
            ),
            fc_errors[0].msg,
        )

    def test_two_proxy_root_with_child_pairs_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne"),
            Crf(show_order=20, model="visit_schedule_app.CrfTwoProxyOne"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne"),
            Crf(show_order=25, model="visit_schedule_app.CrfTwo"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.003", fc_errors[0].id)
        self.assertIn("visit_schedule_app.crfone", fc_errors[0].msg)
        self.assertIn("visit_schedule_app.crftwo", fc_errors[0].msg)
        self.assertIn("visit_schedule_app.crfoneproxyone", fc_errors[0].msg)
        self.assertIn("visit_schedule_app.crftwoproxyone", fc_errors[0].msg)

    def test_two_shared_proxy_root_pairs_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOneProxyOne"),
            Crf(show_order=20, model="visit_schedule_app.CrfTwoProxyOne"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOneProxyTwo"),
            Crf(show_order=25, model="visit_schedule_app.CrfTwoProxyTwo"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxytwo",
                    ],
                    "visit_schedule_app.crftwo": [
                        "visit_schedule_app.crftwoproxyone",
                        "visit_schedule_app.crftwoproxytwo",
                    ],
                }
            ),
            fc_errors[0].msg,
        )

    def test_unscheduled_visit_required_crf_in_prns_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs_unscheduled = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne", required=True),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
            Crf(show_order=30, model="visit_schedule_app.CrfThree"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOne"),
            Crf(show_order=25, model="visit_schedule_app.PrnOne"),
            Crf(show_order=35, model="visit_schedule_app.PrnTwo"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs_unscheduled=crfs_unscheduled,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_unscheduled_visit_required_proxy_root_crf_with_child_proxy_prn_raises(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs_unscheduled = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne", required=True),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne"),
            Crf(show_order=25, model="visit_schedule_app.PrnOne"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs_unscheduled=crfs_unscheduled,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.003", fc_errors[0].id)
        self.assertIn(
            "Proxy root model class appears alongside associated child proxy for a visit.",
            fc_errors[0].msg,
        )
        self.assertIn("proxy_root_model=visit_schedule_app.crfone", fc_errors[0].msg)
        self.assertIn("proxy_model=visit_schedule_app.crfoneproxyone", fc_errors[0].msg)

    def test_unscheduled_visit_crf_proxy_shares_root_true_with_proxy_not_defined_raises(
        self,
    ):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs_unscheduled = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            ),
        )
        prns = CrfCollection(
            Crf(show_order=35, model="visit_schedule_app.CrfOneProxyThree"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs_unscheduled=crfs_unscheduled,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[0].msg,
        )

    def test_missed_visit_required_crf_in_prns_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs_missed = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne", required=True),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
            Crf(show_order=30, model="visit_schedule_app.CrfThree"),
        )
        prns = CrfCollection(
            Crf(show_order=15, model="visit_schedule_app.CrfOne"),
            Crf(show_order=25, model="visit_schedule_app.PrnOne"),
            Crf(show_order=35, model="visit_schedule_app.PrnTwo"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs_missed=crfs_missed,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_missed_visit_required_proxy_root_crf_with_required_child_proxy_crf_raises(
        self,
    ):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs_missed = CrfCollection(
            Crf(show_order=10, model="visit_schedule_app.CrfOne", required=True),
            Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne", required=True),
            Crf(show_order=20, model="visit_schedule_app.CrfTwo"),
            Crf(show_order=30, model="visit_schedule_app.CrfThree"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs_missed=crfs_missed,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.003", fc_errors[0].id)
        self.assertIn("proxy_root_model=visit_schedule_app.crfone", fc_errors[0].msg)
        self.assertIn("proxy_model=visit_schedule_app.crfoneproxyone", fc_errors[0].msg)

    def test_missed_visit_crf_proxy_shares_root_true_with_proxy_not_defined_raises(
        self,
    ):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs_missed = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            )
        )
        prns = CrfCollection(Crf(show_order=35, model="visit_schedule_app.CrfOneProxyThree"))

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs_missed=crfs_missed,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 1)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[0].msg,
        )

    def test_proxy_model_crf_with_identical_proxy_model_in_prns_ok(self):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(Crf(show_order=10, model="visit_schedule_app.CrfOneProxyOne"))
        prns = CrfCollection(Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne"))

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_proxy_shares_root_true_crf_with_identical_proxy_not_sharing_root_in_prns_ok(
        self,
    ):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            )
        )
        prns = CrfCollection(Crf(show_order=15, model="visit_schedule_app.CrfOneProxyOne"))

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_proxy_crf_not_sharing_root_with_identical_proxy_shares_root_true_in_prns_ok(
        self,
    ):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(Crf(show_order=10, model="visit_schedule_app.CrfOneProxyOne"))
        prns = CrfCollection(
            Crf(
                show_order=15,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            )
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertListEqual(fc_errors, [])

    def test_one_crf_proxy_shares_root_true_with_another_proxy_prn_not_defined_raises(
        self,
    ):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            )
        )
        prns = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            ),
            Crf(show_order=35, model="visit_schedule_app.CrfOneProxyThree"),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 3)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[0].msg,
        )
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[1].msg,
        )
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[2].msg,
        )

    def test_multiple_proxy_crfs_prns_some_shared_some_not_raises(
        self,
    ):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            ),
            Crf(
                show_order=35,
                model="visit_schedule_app.CrfOneProxyThree",
            ),
        )
        prns = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
            ),
            Crf(
                show_order=35,
                model="visit_schedule_app.CrfOneProxyThree",
                shares_proxy_root=True,
            ),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        # self.assertEqual(len(fc_errors), 3)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[0].msg,
        )
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[1].msg,
        )
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[2].msg,
        )

    def test_multiple_proxy_crfs_shares_proxy_true_with_proxy_prns_not_defined_raises(
        self,
    ):
        site_visit_schedules._registry = {}
        visit_schedule = self.visit_schedule
        schedule = self.schedule
        crfs = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
                shares_proxy_root=True,
            ),
            Crf(
                show_order=35,
                model="visit_schedule_app.CrfOneProxyThree",
                shares_proxy_root=True,
            ),
        )
        prns = CrfCollection(
            Crf(
                show_order=10,
                model="visit_schedule_app.CrfOneProxyOne",
            ),
            Crf(
                show_order=35,
                model="visit_schedule_app.CrfOneProxyThree",
            ),
        )

        visit = Visit(
            code="1000",
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            facility_name="default",
            crfs=crfs,
            crfs_prn=prns,
            timepoint=1,
        )
        schedule.add_visit(visit)
        visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(visit_schedule)

        vs_errors = visit_schedule_check(app_configs=django_apps.get_app_configs())
        self.assertListEqual(vs_errors, [])

        fc_errors = check_form_collections(app_configs=django_apps.get_app_configs())
        self.assertEqual(len(fc_errors), 3)
        self.assertEqual("edc_visit_schedule.004", fc_errors[0].id)
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[0].msg,
        )
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[1].msg,
        )
        self.assertIn(
            str(
                {
                    "visit_schedule_app.crfone": [
                        "visit_schedule_app.crfoneproxyone",
                        "visit_schedule_app.crfoneproxythree",
                    ]
                }
            ),
            fc_errors[2].msg,
        )



from .schedules_collection import SchedulesCollection, SchedulesCollectionError
from .visit_schedule import (
    AlreadyRegisteredSchedule,
    VisitSchedule,
    VisitScheduleAppointmentModelError,
    VisitScheduleError,
    VisitScheduleNameError,
    VisitScheduleSiteError,
)



from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings, tag
from edc_consent.site_consents import site_consents
from edc_facility.import_holidays import import_holidays
from edc_sites.tests import SiteTestCaseMixin
from edc_utils import get_utcnow

from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from visit_schedule_app.consents import consent_v1
from visit_schedule_app.forms import OffScheduleForm
from visit_schedule_app.models import OnSchedule, SubjectConsent
from visit_schedule_app.visit_schedule import visit_schedule


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
    SITE_ID=30,
)
class TestModels(SiteTestCaseMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        site_visit_schedules.loaded = False
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)
        self.subject_identifier = "1234"
        site_consents.registry = {}
        site_consents.register(consent_v1)

    @tag("2")
    def test_offschedule_ok(self):
        SubjectConsent.objects.create(subject_identifier=self.subject_identifier)
        onschedule = OnSchedule.objects.create(
            subject_identifier=self.subject_identifier,
        )
        data = dict(
            subject_identifier=self.subject_identifier,
            offschedule_datetime=onschedule.onschedule_datetime + relativedelta(months=1),
        )
        form = OffScheduleForm(data=data)
        form.is_valid()



from __future__ import annotations

from collections import Counter, defaultdict
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.checks import Error, Warning
from django.db import models

from .exceptions import SiteVisitScheduleError
from .site_visit_schedules import site_visit_schedules
from .visit import CrfCollection

if TYPE_CHECKING:
    from .schedule import Schedule
    from .visit import Visit
    from .visit_schedule import VisitSchedule


def visit_schedule_check(app_configs, **kwargs):
    errors = []

    if not site_visit_schedules.visit_schedules:
        errors.append(
            Warning("No visit schedules have been registered!", id="edc_visit_schedule.001")
        )
    site_results = check_models()
    for key, results in site_results.items():
        for result in results:
            errors.append(Warning(result, id=f"edc_visit_schedule.{key}"))
    return errors


def check_models() -> dict[str, list]:
    if not site_visit_schedules.loaded:
        raise SiteVisitScheduleError("Registry is not loaded.")
    errors = {"visit_schedules": [], "schedules": [], "visits": []}
    for visit_schedule in site_visit_schedules.visit_schedules.values():
        errors["visit_schedules"].extend(check_visit_schedule_models(visit_schedule))
        for schedule in visit_schedule.schedules.values():
            errors["schedules"].extend(check_schedule_models(schedule))
            for visit in schedule.visits.values():
                errors["visits"].extend(check_visit_models(visit))
    return errors


def check_visit_schedule_models(visit_schedule: VisitSchedule) -> list[str]:
    warnings = []
    for model in ["death_report", "locator", "offstudy"]:
        try:
            getattr(visit_schedule, f"{model}_model_cls")
        except LookupError as e:
            warnings.append(f"{e} See visit schedule '{visit_schedule.name}'.")
    return warnings


def check_schedule_models(schedule: Schedule) -> list[str]:
    warnings = []
    for model in ["onschedule", "offschedule", "appointment"]:
        try:
            getattr(schedule, f"{model}_model_cls")
        except LookupError as e:
            warnings.append(f"{e} See visit schedule '{schedule.name}'.")
    return warnings


def check_visit_models(visit: Visit):
    warnings = []
    models = list(set([f.model for f in visit.all_crfs]))
    for model in models:
        try:
            django_apps.get_model(model)
        except LookupError as e:
            warnings.append(f"{e} Got Visit {visit.code} crf.model={model}.")
    models = list(set([f.model for f in visit.all_requisitions]))
    for model in models:
        try:
            django_apps.get_model(model)
        except LookupError as e:
            warnings.append(f"{e} Got Visit {visit.code} requisition.model={model}.")
    return warnings


def check_form_collections(app_configs, **kwargs):
    errors = []
    for visit_schedule in site_visit_schedules.visit_schedules.values():
        for schedule in visit_schedule.schedules.values():
            for visit in schedule.visits.values():
                for visit_crf_collection, visit_type in [
                    (visit.crfs, "Scheduled"),
                    (visit.crfs_unscheduled, "Unscheduled"),
                    (visit.crfs_missed, "Missed"),
                ]:
                    if proxy_root_alongside_child_err := check_proxy_root_alongside_child(
                        visit=visit,
                        visit_crf_collection=visit_crf_collection,
                        visit_type=visit_type,
                    ):
                        errors.append(proxy_root_alongside_child_err)

                    if same_proxy_root_err := check_multiple_proxies_same_proxy_root(
                        visit=visit,
                        visit_crf_collection=visit_crf_collection,
                        visit_type=visit_type,
                    ):
                        errors.append(same_proxy_root_err)

    return errors


def check_proxy_root_alongside_child(
    visit: Visit,
    visit_crf_collection: CrfCollection,
    visit_type: str,
) -> Error | None:
    all_models = get_models(collection=visit_crf_collection) + get_models(
        collection=visit.crfs_prn
    )
    all_proxy_models = get_proxy_models(collection=visit_crf_collection) + get_proxy_models(
        collection=visit.crfs_prn
    )

    if child_proxies_alongside_proxy_roots := [
        m for m in all_proxy_models if get_proxy_root_model(m) in all_models
    ]:
        proxy_root_child_pairs = [
            (
                f"proxy_root_model={get_proxy_root_model(proxy)._meta.label_lower}",
                f"proxy_model={proxy._meta.label_lower}",
            )
            for proxy in child_proxies_alongside_proxy_roots
        ]
        return Error(
            "Proxy root model class appears alongside associated child "
            "proxy for a visit. "
            f"Got '{visit}' '{visit_type}' visit Crf collection. "
            f"Proxy root:child models: {proxy_root_child_pairs=}",
            id="edc_visit_schedule.003",
        )


def check_multiple_proxies_same_proxy_root(
    visit: Visit,
    visit_crf_collection: CrfCollection,
    visit_type: str,
) -> Error | None:
    # Find all proxy models, and map from their 'proxy root' models
    all_proxy_models = get_proxy_models(collection=visit_crf_collection) + get_proxy_models(
        collection=visit.crfs_prn
    )
    proxy_root_to_child_proxies: defaultdict[str, list[str]] = defaultdict(list)
    for proxy_model in all_proxy_models:
        child_proxy_model_str = proxy_model._meta.label.lower()
        proxy_root_model_str = get_proxy_root_model(proxy_model)._meta.label.lower()
        proxy_root_to_child_proxies[proxy_root_model_str].append(child_proxy_model_str)

    # Find proxy models declared as sharing a proxy root
    proxies_sharing_roots = get_proxy_models(
        collection=CrfCollection(*[f for f in visit_crf_collection if f.shares_proxy_root])
    ) + get_proxy_models(
        collection=CrfCollection(*[f for f in visit.crfs_prn if f.shares_proxy_root])
    )
    proxies_sharing_roots_counter = Counter(
        [m._meta.label.lower() for m in proxies_sharing_roots]
    )

    # Filter out valid models/prepare error list
    for proxy_root in list(proxy_root_to_child_proxies):
        proxies_counter = Counter(proxy_root_to_child_proxies[proxy_root])
        if proxies_counter & proxies_sharing_roots_counter == proxies_counter:
            # OK if proxies counter reflects ALL defined proxy shared roots
            del proxy_root_to_child_proxies[proxy_root]
        elif len(proxies_counter) == 1 and next(iter(proxies_counter.values())) <= 2:
            # OK for a single proxy to be defined in two places (CRFs collection + PRNs)
            del proxy_root_to_child_proxies[proxy_root]
        else:
            proxy_root_to_child_proxies[proxy_root].sort()

    if proxy_root_to_child_proxies:
        return Error(
            "Multiple proxies with same proxy root model appear for "
            "a visit. If this is intentional, consider using "
            "`shares_proxy_root` argument when defining Crf. "
            f"Got '{visit}' '{visit_type}' visit Crf collection. "
            f"Proxy root/child models: {dict(proxy_root_to_child_proxies)}",
            id="edc_visit_schedule.004",
        )


def get_models(collection: CrfCollection) -> list[models.Model]:
    return [f.model_cls for f in collection]


def get_proxy_models(collection: CrfCollection) -> list[models.Model]:
    return [f.model_cls for f in collection if f.model_cls._meta.proxy]


def get_proxy_root_model(proxy_model: models.Model) -> models.Model | None:
    """Returns proxy's root (concrete) model if `proxy_model` is a
    proxy model, else returns None.
    """
    if proxy_model._meta.proxy:
        return proxy_model._meta.concrete_model



from __future__ import annotations

from collections import Counter
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING, Any

from dateutil.relativedelta import relativedelta
from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from edc_utils import floor_secs, formatted_datetime, to_utc
from edc_utils.date import to_local

from .baseline import Baseline
from .exceptions import OffScheduleError, OnScheduleError
from .site_visit_schedules import site_visit_schedules

if TYPE_CHECKING:
    from .model_mixins import OnScheduleModelMixin


def get_default_max_visit_window_gap():
    return getattr(settings, "EDC_VISIT_SCHEDULE_DEFAULT_MAX_VISIT_GAP_ALLOWED", 7)


def get_lower_datetime(instance) -> datetime:
    """Returns the datetime of the lower window"""
    instance.visit.dates.base = to_utc(instance.first.timepoint_datetime)
    return instance.visit.dates.lower


def get_upper_datetime(instance) -> datetime:
    """Returns the datetime of the upper window"""
    instance.visit.dates.base = to_utc(instance.first.timepoint_datetime)
    return instance.visit.dates.upper


def is_baseline(
    instance: Any = None,
    timepoint: Decimal | None = None,
    visit_code_sequence: int | None = None,
    visit_schedule_name: str | None = None,
    schedule_name: str | None = None,
) -> bool:
    return Baseline(
        instance=instance,
        timepoint=timepoint,
        visit_code_sequence=visit_code_sequence,
        visit_schedule_name=visit_schedule_name,
        schedule_name=schedule_name,
    ).value


def raise_if_baseline(subject_visit) -> None:
    if subject_visit and is_baseline(instance=subject_visit):
        raise forms.ValidationError("This form is not available for completion at baseline.")


def raise_if_not_baseline(subject_visit) -> None:
    if subject_visit and not is_baseline(instance=subject_visit):
        raise forms.ValidationError("This form is only available for completion at baseline.")


def get_onschedule_models(
    subject_identifier: str = None, report_datetime: datetime = None
) -> list[str]:
    """Returns a list of onschedule models, in label_lower format,
    for this subject and date.
    """
    onschedule_models = []
    subject_schedule_history_model_cls = django_apps.get_model(
        "edc_visit_schedule.SubjectScheduleHistory"
    )
    for onschedule_model_obj in subject_schedule_history_model_cls.objects.onschedules(
        subject_identifier=subject_identifier, report_datetime=report_datetime
    ):
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            onschedule_model=onschedule_model_obj._meta.label_lower
        )
        onschedule_models.append(schedule.onschedule_model)
    return onschedule_models


def get_offschedule_models(subject_identifier=None, report_datetime=None) -> list[str]:
    """Returns a list of offschedule models, in label_lower format,
    for this subject and date.

    Subject status must be ON_SCHEDULE.

    See also, manager method `onschedules`.
    """
    offschedule_models = []
    subject_schedule_history_model_cls = django_apps.get_model(
        "edc_visit_schedule.SubjectScheduleHistory"
    )
    onschedule_models = subject_schedule_history_model_cls.objects.onschedules(
        subject_identifier=subject_identifier, report_datetime=report_datetime
    )
    for onschedule_model_obj in onschedule_models:
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            onschedule_model=onschedule_model_obj._meta.label_lower
        )
        offschedule_models.append(schedule.offschedule_model)
    return offschedule_models


def off_schedule_or_raise(
    subject_identifier=None,
    report_datetime=None,
    visit_schedule_name=None,
    schedule_name=None,
) -> None:
    """Returns True if subject is on the given schedule
    on this date.
    """
    visit_schedule = site_visit_schedules.get_visit_schedule(
        visit_schedule_name=visit_schedule_name
    )
    schedule = visit_schedule.schedules.get(schedule_name)
    if schedule.is_onschedule(subject_identifier, report_datetime):
        raise OnScheduleError(
            f"Not allowed. Subject {subject_identifier} is on schedule "
            f"{visit_schedule.verbose_name}.{schedule_name} on "
            f"{formatted_datetime(report_datetime)}. "
            f"See model '{schedule.offschedule_model_cls().verbose_name}'"
        )


def off_all_schedules_or_raise(subject_identifier: str = None):
    """Raises an exception if subject is still on any schedule."""
    for visit_schedule in site_visit_schedules.get_visit_schedules().values():
        for schedule in visit_schedule.schedules.values():
            try:
                with transaction.atomic():
                    schedule.onschedule_model_cls.objects.get(
                        subject_identifier=subject_identifier
                    )
            except ObjectDoesNotExist:
                pass
            else:
                try:
                    with transaction.atomic():
                        schedule.offschedule_model_cls.objects.get(
                            subject_identifier=subject_identifier
                        )
                except ObjectDoesNotExist:
                    model_name = schedule.offschedule_model_cls()._meta.verbose_name.title()
                    raise OffScheduleError(
                        f"Subject cannot be taken off study. Subject is still on a "
                        f"schedule. Got schedule '{visit_schedule.name}.{schedule.name}. "
                        f"Complete the offschedule form `{model_name}` first. "
                        f"Subject identifier='{subject_identifier}', "
                    )
    return True


def offstudy_datetime_after_all_offschedule_datetimes(
    subject_identifier: str = None,
    offstudy_datetime: datetime = None,
    exception_cls=None,
) -> None:
    exception_cls = exception_cls or forms.ValidationError
    for visit_schedule in site_visit_schedules.get_visit_schedules().values():
        for schedule in visit_schedule.schedules.values():
            try:
                schedule.onschedule_model_cls.objects.get(
                    subject_identifier=subject_identifier
                )
            except ObjectDoesNotExist:
                pass
            else:
                try:
                    offschedule_obj = schedule.offschedule_model_cls.objects.get(
                        subject_identifier=subject_identifier,
                        offschedule_datetime__gt=offstudy_datetime,
                    )
                except ObjectDoesNotExist:
                    pass
                else:
                    offschedule_datetime = formatted_datetime(
                        offschedule_obj.offschedule_datetime
                    )
                    raise exception_cls(
                        "`Offstudy` datetime cannot be before any `offschedule` datetime. "
                        f"Got {subject_identifier} went off schedule "
                        f"`{visit_schedule.name}.{schedule.name}` on "
                        f"{offschedule_datetime}."
                    )


def report_datetime_within_onschedule_offschedule_datetimes(
    subject_identifier: str = None,
    report_datetime: datetime = None,
    visit_schedule_name: str = None,
    schedule_name: str = None,
    exception_cls=None,
):
    exception_cls = exception_cls or forms.ValidationError
    visit_schedule = site_visit_schedules.get_visit_schedule(visit_schedule_name)
    schedule = visit_schedule.schedules.get(schedule_name)
    try:
        onschedule_obj = schedule.onschedule_model_cls.objects.get(
            subject_identifier=subject_identifier
        )
    except ObjectDoesNotExist:
        raise OnScheduleError(
            f"Subject is not on schedule. {visit_schedule_name}.{schedule_name}. "
            f"Got {subject_identifier}"
        )
    try:
        offschedule_obj = schedule.offschedule_model_cls.objects.get(
            subject_identifier=subject_identifier,
            offschedule_datetime__lte=report_datetime,
        )
    except ObjectDoesNotExist:
        offschedule_obj = None
        offschedule_datetime = report_datetime
    else:
        offschedule_datetime = offschedule_obj.offschedule_datetime
    if not (
        floor_secs(onschedule_obj.onschedule_datetime)
        <= floor_secs(report_datetime)
        <= floor_secs(offschedule_datetime)
    ):
        onschedule_datetime = formatted_datetime(onschedule_obj.onschedule_datetime)
        if offschedule_obj:
            offschedule_datetime = formatted_datetime(offschedule_obj.offschedule_datetime)
            error_msg = (
                "Invalid report datetime. Expected a datetime between "
                f"{onschedule_datetime} and {offschedule_datetime}. "
                "See onschedule and offschedule."
            )
        else:
            error_msg = (
                "Invalid report datetime. Expected a datetime on or after "
                f"{onschedule_datetime}. See onschedule."
            )
        raise exception_cls(error_msg)


def get_onschedule_model_instance(
    subject_identifier: str,
    reference_datetime: datetime,
    visit_schedule_name: str,
    schedule_name: str,
) -> OnScheduleModelMixin:
    """Returns the onschedule model instance

    Increment reference_datetime by 1 sec to avoid millisecond
    in lte comparison.
    """
    schedule = site_visit_schedules.get_visit_schedule(visit_schedule_name).schedules.get(
        schedule_name
    )
    model_cls = django_apps.get_model(schedule.onschedule_model)
    try:
        onschedule_obj = model_cls.objects.get(
            subject_identifier=subject_identifier,
            onschedule_datetime__lte=to_utc(reference_datetime) + relativedelta(seconds=1),
        )
    except ObjectDoesNotExist as e:
        dte_as_str = formatted_datetime(to_local(reference_datetime))
        raise OffScheduleError(
            "Subject is not on a schedule. Using subject_identifier="
            f"`{subject_identifier}` and appt_datetime=`{dte_as_str}`. Got {e}"
        )
    return onschedule_obj


def get_duplicates(list_items: list[Any]) -> list:
    return [n for n, count in Counter(list_items).items() if count > 1]



import sys

from django.apps.config import AppConfig as DjangoAppConfig
from django.core.checks import register
from django.core.management.color import color_style
from django.db.models.signals import post_migrate

from .site_visit_schedules import site_visit_schedules
from .system_checks import check_form_collections, visit_schedule_check

style = color_style()


def populate_visit_schedule(sender=None, **kwargs):
    from .models import VisitSchedule

    sys.stdout.write(style.MIGRATE_HEADING("Populating visit schedule:\n"))
    VisitSchedule.objects.update(active=False)
    site_visit_schedules.to_model(VisitSchedule)
    sys.stdout.write("Done.\n")
    sys.stdout.flush()


class AppConfig(DjangoAppConfig):
    name = "edc_visit_schedule"
    verbose_name = "Visit Schedules"
    validate_models = True
    include_in_administration_section = True

    def ready(self):
        post_migrate.connect(populate_visit_schedule, sender=self)
        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        site_visit_schedules.autodiscover()
        sys.stdout.write(f" Done loading {self.verbose_name}.\n")
        register(visit_schedule_check)
        register(check_form_collections)



from django.apps import AppConfig as DjangoAppConfig
from edc_metadata.apps import AppConfig as BaseEdcMetadataAppConfig


class AppConfig(DjangoAppConfig):
    name = "visit_schedule_app"


class EdcMetadataAppConfig(BaseEdcMetadataAppConfig):
    reason_field = {"edc_visit_tracking.subjectvisit": "reason"}



from edc_model.models import BaseUuidModel
from edc_sites.model_mixins import SiteModelMixin

from ..model_mixins import OffScheduleModelMixin


class OffSchedule(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):

    """A model used by the system. Records a subject as no longer on
    a schedule.
    """

    class Meta(OffScheduleModelMixin.Meta, BaseUuidModel.Meta):
        pass



from django.test import TestCase

from edc_visit_schedule.visit import Crf, CrfCollection, FormsCollectionError


class TestCrfCollection(TestCase):
    def test_crf_collection_ok(self):
        crfs = [
            Crf(show_order=100, model="x.one"),
            Crf(show_order=200, model="x.two"),
            Crf(show_order=300, model="x.three"),
        ]
        try:
            CrfCollection(*crfs)
        except FormsCollectionError:
            self.fail("FormsCollectionError unexpectedly raised")

    def test_crf_collection_with_duplicate_ordering_raises(self):
        crfs = [
            Crf(show_order=100, model="x.one"),
            Crf(show_order=200, model="x.two"),
            Crf(show_order=100, model="x.three"),
        ]
        with self.assertRaises(FormsCollectionError) as cm:
            CrfCollection(*crfs)
        self.assertIn(
            'CrfCollection "show order" must be a unique sequence.',
            str(cm.exception),
        )
        self.assertIn("Duplicates [100].", str(cm.exception))

    def test_crf_collection_with_duplicate_models_raises(self):
        crfs = [
            Crf(show_order=100, model="x.one"),
            Crf(show_order=200, model="x.two"),
            Crf(show_order=300, model="x.one"),
        ]
        with self.assertRaises(FormsCollectionError) as cm:
            CrfCollection(*crfs)
        self.assertIn("Expected to be a unique sequence of crf/models.", str(cm.exception)),
        self.assertIn(" Duplicates ['x.one'].", str(cm.exception))



from datetime import timedelta

from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_utils import get_utcnow

from edc_visit_schedule.schedule import AlreadyRegisteredVisit, Schedule
from edc_visit_schedule.schedule.schedule import ScheduleNameError, VisitTimepointError
from edc_visit_schedule.schedule.visit_collection import VisitCollectionError
from edc_visit_schedule.system_checks import check_schedule_models
from edc_visit_schedule.visit import Visit
from visit_schedule_app.consents import consent_v1
from visit_schedule_app.models import OffSchedule, OnSchedule


class TestSchedule(TestCase):
    def test_schedule_name(self):
        self.assertRaises(ScheduleNameError, Schedule, name="sched  ule")

    def test_visit_schedule_repr(self):
        """Asserts repr evaluates correctly."""
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        self.assertTrue(schedule.__repr__())

    def test_visit_schedule_field_value(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        self.assertEqual(schedule.field_value, "schedule")

    def test_schedule_bad_label_lower(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="x.x",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        errors = check_schedule_models(schedule)
        self.assertIsNotNone(errors)

    def test_schedule_bad_label_lower2(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="x.x",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        errors = check_schedule_models(schedule)
        self.assertIsNotNone(errors)

    def test_schedule_onschedule_model_cls(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        self.assertEqual(schedule.onschedule_model_cls, OnSchedule)

    def test_schedule_offschedule_model_cls(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        self.assertEqual(schedule.offschedule_model_cls, OffSchedule)

    def test_schedule_ok(self):
        Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )

    def test_add_visits(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        for i in range(0, 5):
            visit = Visit(
                code=str(i),
                timepoint=i,
                rbase=relativedelta(days=i),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
            )
            try:
                schedule.add_visit(visit=visit)
            except AlreadyRegisteredVisit as e:
                self.fail(f"Exception unexpectedly raised. Got {e}")

    def test_add_visits_duplicate_code(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        visit = Visit(
            code=str(0),
            title="erik0",
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        schedule.add_visit(visit=visit)
        visit = Visit(
            code=str(0),
            title="erik1",
            timepoint=1,
            rbase=relativedelta(days=1),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        self.assertRaises(AlreadyRegisteredVisit, schedule.add_visit, visit=visit)

    def test_add_visits_duplicate_title(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        visit = Visit(
            code=str(0),
            title="erik",
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        schedule.add_visit(visit=visit)
        visit = Visit(
            code=str(1),
            title="erik",
            timepoint=1,
            rbase=relativedelta(days=1),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        self.assertRaises(AlreadyRegisteredVisit, schedule.add_visit, visit=visit)

    def test_add_visits_custom_base_timepoint(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
            base_timepoint=1,
        )
        visit = Visit(
            code=str(0),
            title="erik",
            timepoint=1,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        try:
            schedule.add_visit(visit=visit)
        except VisitTimepointError:
            self.fail("VisitTimepointError unexpectedly raised")

    def test_add_visits_base_timepoint_mismatch(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
            base_timepoint=1,
        )
        visit = Visit(
            code=str(0),
            title="erik",
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        self.assertRaises(VisitTimepointError, schedule.add_visit, visit=visit)

    def test_add_visits_duplicate_timepoint(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        visit = Visit(
            code=str(0),
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        schedule.add_visit(visit=visit)
        visit = Visit(
            code=str(1),
            timepoint=0,
            rbase=relativedelta(days=1),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        self.assertRaises(AlreadyRegisteredVisit, schedule.add_visit, visit=visit)

    def test_add_visits_duplicate_rbase(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )
        visit = Visit(
            code=str(0),
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        schedule.add_visit(visit=visit)
        visit = Visit(
            code=str(1),
            timepoint=1,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        self.assertRaises(AlreadyRegisteredVisit, schedule.add_visit, visit=visit)


class TestScheduleWithVisits(TestCase):
    def setUp(self):
        self.schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
        )

    def test_wont_accept_visit_before_base_timepoint(self):
        schedule = Schedule(
            name="schedule",
            onschedule_model="visit_schedule_app.onschedule",
            offschedule_model="visit_schedule_app.offschedule",
            consent_definitions=[consent_v1],
            appointment_model="edc_appointment.appointment",
            base_timepoint=1,
        )
        visit = Visit(
            code=str(1),
            timepoint=1,
            rbase=relativedelta(days=1),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        schedule.add_visit(visit=visit)

        visit = Visit(
            code=str(0),
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        self.assertRaises(VisitTimepointError, schedule.add_visit, visit=visit)

    def test_first_visit_added_must_be_base(self):
        visits = []
        for i in [3, 0]:
            visits.append(
                Visit(
                    code=str(i),
                    timepoint=i,
                    rbase=relativedelta(days=i),
                    rlower=relativedelta(days=0),
                    rupper=relativedelta(days=6),
                )
            )
        # attempt to add visit where timepoint == 3 - raises
        self.assertRaises(VisitTimepointError, self.schedule.add_visit, visit=visits[0])
        # attempt to add visit where timepoint == 0 - ok
        try:
            self.schedule.add_visit(visit=visits[1])
        except VisitTimepointError:
            self.fail("VisitTimepointError unexpectedly raised")
        # attempt to add visit where timepoint == 3 again - ok
        try:
            self.schedule.add_visit(visit=visits[0])
        except VisitTimepointError:
            self.fail("VisitTimepointError unexpectedly raised")

    def test_order(self):
        for i in [0, 3, 5, 1, 2, 4]:
            visit = Visit(
                code=str(i),
                timepoint=i,
                rbase=relativedelta(days=i),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
            )
            self.schedule.add_visit(visit=visit)
        self.assertEqual(
            [v.timepoint for v in self.schedule.visits.values()], [0, 1, 2, 3, 4, 5]
        )

    def test_first_visit(self):
        first_visit = Visit(
            code=str(0),
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        self.schedule.add_visit(visit=first_visit)
        for i in range(1, 5):
            visit = Visit(
                code=str(i),
                timepoint=i,
                rbase=relativedelta(days=i),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
            )
            self.schedule.add_visit(visit=visit)
        self.assertEqual(self.schedule.visits.first, first_visit)

    def test_last_visit(self):
        for i in range(0, 5):
            visit = Visit(
                code=str(i),
                timepoint=i,
                rbase=relativedelta(days=i),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
            )
            self.schedule.add_visit(visit=visit)
        visit = Visit(
            code=str(6),
            timepoint=6,
            rbase=relativedelta(days=6),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
        )
        self.schedule.add_visit(visit=visit)
        self.assertEqual(self.schedule.visits.last, visit)

    def test_get_visit(self):
        for i in range(0, 5):
            visit = Visit(
                code=str(i),
                timepoint=i,
                rbase=relativedelta(days=i),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
            )
            self.schedule.add_visit(visit=visit)
        visit = self.schedule.visits.get("3")
        self.assertEqual(visit.code, "3")
        self.assertRaises(VisitCollectionError, self.schedule.visits.get, "BLAH")

    def test_previous_visit(self):
        for i in range(0, 5):
            visit = Visit(
                code=str(i),
                timepoint=i,
                rbase=relativedelta(days=i),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
            )
            self.schedule.add_visit(visit=visit)
        visit = self.schedule.visits.previous("3")
        self.assertEqual(visit.code, "2")

    def test_next_visit(self):
        for i in range(0, 5):
            visit = Visit(
                code=str(i),
                timepoint=i,
                rbase=relativedelta(days=i),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
            )
            self.schedule.add_visit(visit=visit)
        visit = self.schedule.visits.next("3")
        self.assertEqual(visit.code, "4")

    def test_previous_visit_none(self):
        for i in range(0, 5):
            visit = Visit(
                code=str(i),
                timepoint=i,
                rbase=relativedelta(days=i),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
            )
            self.schedule.add_visit(visit=visit)
        self.assertIsNone(self.schedule.visits.previous("0"))

    def test_next_visit_none(self):
        for i in range(0, 5):
            visit = Visit(
                code=str(i),
                timepoint=i,
                rbase=relativedelta(days=i),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
            )
            self.schedule.add_visit(visit=visit)
        self.assertIsNone(self.schedule.visits.next("5"))

    def test_visit_dates(self):
        dt = get_utcnow()
        for index, seq in enumerate(range(0, 5)):
            visit = Visit(
                code=str(seq),
                timepoint=seq * (index + 1),
                rbase=relativedelta(days=seq * (index + 1)),
                rlower=relativedelta(days=0),
                rupper=relativedelta(days=6),
            )
            self.schedule.add_visit(visit=visit)
        index = 0
        for k, v in self.schedule.visits.timepoint_dates(dt=dt).items():
            self.assertEqual(v - dt, timedelta(index * (index + 1)), msg=k)
            index += 1



from __future__ import annotations

from collections import namedtuple
from datetime import datetime
from decimal import Decimal
from typing import Tuple
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from edc_utils import to_utc


class WindowPeriod:
    def __init__(
        self,
        rlower: relativedelta = None,
        rupper: relativedelta = None,
        timepoint: Decimal | None = None,
        base_timepoint: Decimal | None = None,
        no_floor: bool | None = None,
        no_ceil: bool | None = None,
    ):
        self.rlower = rlower
        self.rupper = rupper
        self.no_floor = no_floor
        self.no_ceil = no_ceil
        self.timepoint = Decimal("0.0") if timepoint is None else timepoint
        base_timepoint = Decimal("0.0") if base_timepoint is None else base_timepoint
        if self.timepoint == base_timepoint:
            self.no_floor = True

    def get_window(self, dt=None) -> Tuple[datetime, datetime]:
        """Returns a tuple of the lower and upper datetimes in UTC."""

        dt_floor = (
            to_utc(dt)
            if self.no_floor
            else dt.replace(hour=0, minute=0, second=0, microsecond=0).astimezone(
                ZoneInfo("UTC")
            )
        )
        dt_ceil = (
            to_utc(dt)
            if self.no_ceil
            else dt.replace(hour=23, minute=59, second=59, microsecond=999999).astimezone(
                ZoneInfo("UTC")
            )
        )
        window = namedtuple("Window", ["lower", "upper"])
        return window(dt_floor - self.rlower, dt_ceil + self.rupper)



from .offschedule import OffSchedule
from .onschedule import OnSchedule
from .signals import (
    offschedule_model_on_post_save,
    onschedule_model_on_post_save,
    put_subject_on_schedule_on_post_save,
)
from .subject_schedule_history import SubjectScheduleHistory
from .visit_schedule import VisitSchedule
